@@ -123,7 +123,7 @@ KIOGDrive::Action KIOGDrive::handleError(const KGAPI2::Job &job, const QUrl &url
             error(KIO::ERR_DISK_FULL, url.toDisplayString());
             return Fail;
         default:
-            error(KIO::ERR_SLAVE_DEFINED, job.errorString());
+            error(KIO::ERR_WORKER_DEFINED, job.errorString());
             return Fail;
     }
 
@@ -229,6 +229,9 @@ KIO::UDSEntry KIOGDrive::fileToUDSEntry(const FilePtr &origFile, const QString &
     entry.fastInsert(GDriveUDSEntryExtras::Md5, file->md5Checksum());
     entry.fastInsert(GDriveUDSEntryExtras::LastModifyingUser, file->lastModifyingUserName());
     entry.fastInsert(GDriveUDSEntryExtras::Owners, file->ownerNames().join(QStringLiteral(", ")));
+    if (file->sharedWithMeDate().isValid()) {
+        entry.fastInsert(GDriveUDSEntryExtras::SharedWithMeDate, QLocale::system().toString(file->sharedWithMeDate(), QLocale::LongFormat));
+    }
 
     return entry;
 }
@@ -264,6 +267,19 @@ KIO::UDSEntry KIOGDrive::newAccountUDSEntry()
     return entry;
 }
 
+KIO::UDSEntry KIOGDrive::sharedWithMeUDSEntry()
+{
+    KIO::UDSEntry entry;
+
+    entry.fastInsert(KIO::UDSEntry::UDS_NAME, GDriveUrl::SharedWithMeDir);
+    entry.fastInsert(KIO::UDSEntry::UDS_DISPLAY_NAME, i18nc("folder containing gdrive files shared with me", "Shared With Me"));
+    entry.fastInsert(KIO::UDSEntry::UDS_FILE_TYPE, S_IFDIR);
+    entry.fastInsert(KIO::UDSEntry::UDS_ICON_NAME, QStringLiteral("folder-publicshare"));
+    entry.fastInsert(KIO::UDSEntry::UDS_ACCESS, S_IRUSR);
+
+    return entry;
+}
+
 KIO::UDSEntry KIOGDrive::accountToUDSEntry(const QString &accountNAme)
 {
     KIO::UDSEntry entry;
@@ -310,7 +326,7 @@ void KIOGDrive::createAccount()
     }
 
     if (m_accountManager->accounts().isEmpty()) {
-        error(KIO::ERR_SLAVE_DEFINED, i18n("There are no Google Drive accounts enabled. Please add at least one."));
+        error(KIO::ERR_WORKER_DEFINED, i18n("There are no Google Drive accounts enabled. Please add at least one."));
         return;
     }
 
@@ -446,7 +462,7 @@ KIO::UDSEntry KIOGDrive::fetchSharedDrivesRootEntry(const QString &accountId, Fe
             canCreateDrives = about->canCreateDrives();
         }
     }
-    qCDebug(GDRIVE) << "Account" << accountId << (canCreateDrives ? "can" : "can't") << "create Drives";
+    qCDebug(GDRIVE) << "Account" << accountId << (canCreateDrives ? "can" : "can't") << "create Shared Drives";
 
     KIO::UDSEntry entry;
 
@@ -517,7 +533,7 @@ QString KIOGDrive::resolveFileIdFromPath(const QString &path, PathFlags flags)
     const auto gdriveUrl = GDriveUrl(url);
     Q_ASSERT(!gdriveUrl.isRoot());
 
-    if (gdriveUrl.isAccountRoot() || gdriveUrl.isTrashDir()) {
+    if (gdriveUrl.isAccountRoot() || gdriveUrl.isTrashDir() || gdriveUrl.isSharedWithMeRoot()) {
         qCDebug(GDRIVE) << "Resolved" << path << "to account root";
         return rootFolderId(gdriveUrl.account());
     }
@@ -535,16 +551,19 @@ QString KIOGDrive::resolveFileIdFromPath(const QString &path, PathFlags flags)
         return QString();
     }
 
-    // Try to recursively resolve ID of parent path - either from cache, or by
-    // querying Google
-    const QString parentId = resolveFileIdFromPath(gdriveUrl.parentPath(), KIOGDrive::PathIsFolder);
-    if (parentId.isEmpty()) {
-        // We failed to resolve parent -> error
-        return QString();
+    QString parentId;
+    if (!gdriveUrl.isSharedWithMeTopLevel()) {
+        // Try to recursively resolve ID of parent path - either from cache, or by querying Google
+        parentId = resolveFileIdFromPath(gdriveUrl.parentPath(), KIOGDrive::PathIsFolder);
+        if (parentId.isEmpty()) {
+            // We failed to resolve parent -> error
+            return QString();
+        }
+        qCDebug(GDRIVE) << "Getting ID for" << gdriveUrl.filename() << "in parent with ID" << parentId;
+    } else {
+        qCDebug(GDRIVE) << "Getting ID for" << gdriveUrl.filename() << "(top-level shared-with-me file without a parentId)";
     }
 
-    qCDebug(GDRIVE) << "Getting ID for" << gdriveUrl.filename() << "in parent with ID" << parentId;
-
     FileSearchQuery query;
     if (flags != KIOGDrive::None) {
         query.addQuery(FileSearchQuery::MimeType,
@@ -552,7 +571,9 @@ QString KIOGDrive::resolveFileIdFromPath(const QString &path, PathFlags flags)
                        GDriveHelper::folderMimeType());
     }
     query.addQuery(FileSearchQuery::Title, FileSearchQuery::Equals, gdriveUrl.filename());
-    query.addQuery(FileSearchQuery::Parents, FileSearchQuery::In, parentId);
+    if (!parentId.isEmpty()) {
+        query.addQuery(FileSearchQuery::Parents, FileSearchQuery::In, parentId);
+    }
     query.addQuery(FileSearchQuery::Trashed, FileSearchQuery::Equals, gdriveUrl.isTrashed());
 
     const QString accountId = gdriveUrl.account();
@@ -700,7 +721,7 @@ void KIOGDrive::listDir(const QUrl &url)
     const auto account = getAccount(accountId);
     if (account->accountName().isEmpty()) {
         qCDebug(GDRIVE) << "Unknown account" << accountId << "for" << url;
-        error(KIO::ERR_SLAVE_DEFINED, i18n("%1 isn't a known GDrive account", accountId));
+        error(KIO::ERR_WORKER_DEFINED, i18n("%1 isn't a known GDrive account", accountId));
         return;
     }
 
@@ -709,6 +730,9 @@ void KIOGDrive::listDir(const QUrl &url)
         auto entry = fetchSharedDrivesRootEntry(accountId);
         listEntry(entry);
         folderId = rootFolderId(accountId);
+
+        auto sharedWithMeEntry = sharedWithMeUDSEntry();
+        listEntry(sharedWithMeEntry);
     } else if (gdriveUrl.isSharedDrivesRoot()) {
         listSharedDrivesRoot(url);
         return;
@@ -726,7 +750,12 @@ void KIOGDrive::listDir(const QUrl &url)
 
     FileSearchQuery query;
     query.addQuery(FileSearchQuery::Trashed, FileSearchQuery::Equals, false);
-    query.addQuery(FileSearchQuery::Parents, FileSearchQuery::In, folderId);
+    // Top-level Shared-with-me files don't have a parent.
+    if (gdriveUrl.isSharedWithMeRoot()) {
+        query.addQuery(FileSearchQuery::SharedWithMe, FileSearchQuery::Equals, true);
+    } else {
+        query.addQuery(FileSearchQuery::Parents, FileSearchQuery::In, folderId);
+    }
     FileFetchJob fileFetchJob(query, getAccount(accountId));
     const auto extraFields =
         QStringList({ KGAPI2::Drive::File::Fields::Labels,
@@ -833,6 +862,14 @@ void KIOGDrive::stat(const QUrl &url)
         return;
     }
 
+    if (gdriveUrl.isSharedWithMeRoot()) {
+        qCDebug(GDRIVE) << "stat()ing Shared With Me path";
+        const KIO::UDSEntry entry = sharedWithMeUDSEntry();
+        statEntry(entry);
+        finished();
+        return;
+    }
+
     // We are committed to stat()ing an url that belongs to
     // an account (i.e. not root or new account path), lets
     // make sure we know about the account
@@ -840,7 +877,7 @@ void KIOGDrive::stat(const QUrl &url)
     const auto account = getAccount(accountId);
     if (account->accountName().isEmpty()) {
         qCDebug(GDRIVE) << "Unknown account" << accountId << "for" << url;
-        error(KIO::ERR_SLAVE_DEFINED, i18n("%1 isn't a known GDrive account", accountId));
+        error(KIO::ERR_WORKER_DEFINED, i18n("%1 isn't a known GDrive account", accountId));
         return;
     }
 
@@ -945,18 +982,23 @@ void KIOGDrive::get(const QUrl &url)
     if (GDriveHelper::isGDocsDocument(file)) {
         downloadUrl = GDriveHelper::convertFromGDocs(file);
     } else {
-        downloadUrl = file->downloadUrl();
+        downloadUrl = GDriveHelper::downloadUrl(file);
     }
 
     mimeType(file->mimeType());
 
     FileFetchContentJob contentJob(downloadUrl, getAccount(accountId));
+    QObject::connect(&contentJob, &KGAPI2::Job::progress, [this](KGAPI2::Job *, int processed, int total) {
+        processedSize(processed);
+        totalSize(total);
+    });
     if (!runJob(contentJob, url, accountId)) {
         return;
     }
 
     QByteArray contentData = contentJob.data();
 
+    processedSize(contentData.size());
     totalSize(contentData.size());
 
     // data() has a maximum transfer size of 14 MiB so we need to send it in chunks.
@@ -1016,6 +1058,13 @@ bool KIOGDrive::readPutData(QTemporaryFile &tempFile, FilePtr &fileMetaData)
 
 bool KIOGDrive::runJob(KGAPI2::Job &job, const QUrl &url, const QString &accountId)
 {
+    auto account = getAccount(accountId);
+    if (account->accessToken().isEmpty()) {
+        qCWarning(GDRIVE) << "Expired or missing access/refresh token for account" << accountId;
+        error(KIO::ERR_WORKER_DEFINED, i18n("Expired or missing access tokens for account %1", accountId));
+        return false;
+    }
+
     KIOGDrive::Action action = KIOGDrive::Fail;
     Q_FOREVER {
         qCDebug(GDRIVE) << "Running job" << (&job) << "with accessToken" << GDriveHelper::elideToken(job.account()->accessToken());
@@ -1029,7 +1078,7 @@ bool KIOGDrive::runJob(KGAPI2::Job &job, const QUrl &url, const QString &account
         } else if (action == KIOGDrive::Fail) {
             return false;
         }
-        job.setAccount(getAccount(accountId));
+        job.setAccount(account);
         job.restart();
     };
 
@@ -1260,7 +1309,7 @@ void KIOGDrive::del(const QUrl &url, bool isfile)
     // Trying to delete the Team Drive root is pointless
     if (gdriveUrl.isSharedDrivesRoot()) {
         qCDebug(GDRIVE) << "Tried deleting Shared Drives root.";
-        error(KIO::ERR_SLAVE_DEFINED, i18n("Can't delete Shared Drives root."));
+        error(KIO::ERR_WORKER_DEFINED, i18n("Can't delete Shared Drives root."));
         return;
     }
 
